# -*- coding: utf-8 -*-
"""
Provides a set of global variables within the orkgnlp_context.
"""

import os

from orkgnlp.common.util import io
from orkgnlp.common.util.datastructure import StrictDict

_current_dir = os.path.dirname(os.path.realpath(__file__))

"""
Context dictionary for library's global configuration variables.
Keys represent environment variables and values represent default values.
Values can be changed using the corresponding setter function.
"""
orkgnlp_context = StrictDict(
    {
        "ORKG_NLP_DATA_CACHE_ROOT": os.path.join(os.path.expanduser("~"), "orkgnlp_data"),
        "ORKG_NLP_VERBOSITY": True,
        "HUGGINGFACE_REPOS": io.read_json(
            os.path.join(_current_dir, "..", "..", "huggingface_repos.json")
        ),
        "SERVICE_MAP": {},
    }
)


def set_data_cache_root(cache_root: str):
    """
    Overrides the value of ORKG_NLP_DATA_CACHE_ROOT. See :doc:`../configure`

    :param cache_root: Path to the data cache root. The path must be absolute.
    """
    orkgnlp_context["ORKG_NLP_DATA_CACHE_ROOT"] = cache_root


def set_verbosity(verbose: bool):
    """
    Overrides the value of ORKG_NLP_VERBOSITY. See :doc:`../configure`

    :param verbose: Indicates whether orkgnlp is in verbose mode.
    """
    orkgnlp_context["ORKG_NLP_VERBOSITY"] = verbose



# -*- coding: utf-8 -*-
""" Templates Recommendation service decoder. """
from typing import Any, Dict, Generator, List

from overrides import overrides

from orkgnlp.common.service.base import ORKGNLPBaseDecoder


class TemplatesRecommenderDecoder(ORKGNLPBaseDecoder):
    """
    The TemplatesRecommenderDecoder decodes the Templates Recommendation service model's output
    to a user-friendly one.
    """

    def __init__(self, templates: List[Dict[str, str]]):
        """

        :param templates: templates used for training the service models as premises.
        """
        super().__init__()

        self.templates: List[Dict[str, str]] = templates
        self.id2label = {"0": "entailment", "1": "contradiction", "2": "neutral"}

    @overrides(check_signature=False)
    def decode(self, model_output: Generator[Any, None, None], top_n: int, **kwargs: Any) -> Any:
        templates = []
        for idx, prediction in enumerate(model_output):
            class_id = prediction.logits.argmax(dim=-1).item()
            score = prediction.logits[0][class_id].item()

            if self.id2label[str(class_id)] == "entailment":
                templates.append(
                    {
                        "id": self.templates[idx]["id"],
                        "label": self.templates[idx]["label"],
                        "score": score,
                    }
                )

        templates = sorted(templates, key=lambda i: i["score"], reverse=True)

        return templates[:top_n]



# -*- coding: utf-8 -*-
""" Base interfaces. """
import os
from typing import Any, Dict, Generator, Tuple, Union

from overrides import EnforceOverrides

from orkgnlp.common.config import orkgnlp_context
from orkgnlp.common.tools import downloader
from orkgnlp.common.util.decorators import singleton
from orkgnlp.common.util.exceptions import (
    ORKGNLPIllegalStateError,
    ORKGNLPValidationError,
)


class PipelineExecutorComponent:
    """
    The PipelineExecutorComponent represents a component of a PipelineExecutor
    """

    def release_memory(self):
        """
        Releases the memory of all available attributes in a pipeline component.
        """
        attributes = list(self.__dict__.keys())
        for attribute in attributes:
            delattr(self, attribute)


class ORKGNLPBaseEncoder(EnforceOverrides, PipelineExecutorComponent):
    """
    The ORKGNLPBaseEncoder is  the base encoder class. You can freely inherit this class, implement its
    ``encode(raw_input, **kwargs)`` function and use it to encode your user input to a model-friendly format.

    Using the ORKGNLPBaseEncoder as your service encoder results in passing the same user's input to the model.
    """

    def __init__(self):
        pass

    def encode(self, raw_input: Any, **kwargs: Any) -> Tuple[Any, Dict[str, Any]]:
        """
        Encodes the ``raw_input`` to a model-friendly format.

        :param raw_input: The user's input to be encoded.
        :return: The model-friendly output and kwargs.
        """
        return (raw_input,), kwargs


class ORKGNLPBaseDecoder(EnforceOverrides, PipelineExecutorComponent):
    """
    The ORKGNLPBaseDecoder is  the base decoder class. You can freely inherit this class, implement its
    ``decode(model_output, **kwargs)`` function and use it to decode your model output to a user-friendly format.

    Using the ORKGNLPBaseDecoder as your service decoder results in returning the same model's output to the user.
    """

    def __init__(self):
        pass

    def decode(self, model_output: Union[Any, Generator[Any, None, None]], **kwargs: Any) -> Any:
        """
        Decodes the model's ``output`` to a user-friendly format.

        :param model_output: The model's output to be decoded.
        :return: The user-friendly output.
        """
        return model_output, kwargs


class ORKGNLPBaseRunner(EnforceOverrides, PipelineExecutorComponent):
    """
    The ORKGNLPBaseRunner is  the base runner class. It requires a model object while initialization.
    This runner must be inherited and the `run(*args, **kwargs)` must be overridden,
    thus running this runner raises an ``NotImplementedError``.
    """

    def __init__(self, model: Any):
        """

        :param model: The model to be run.
        :type model: Model object. See the inheriting classes for further information.
        """
        self._model: Any = model

    def run(self, *args: Any, **kwargs: Any):
        """

        :raise: NotImplementedError
        """
        raise NotImplementedError("Subclass must implement abstract method")


class PipelineExecutor:
    """
    The PipelineExecutor executes a full service workflow given its encoder, runner and decoder.
    See the ``run`` function description for further information.
    """

    def __init__(
        self,
        encoder: ORKGNLPBaseEncoder,
        runner: ORKGNLPBaseRunner,
        decoder: ORKGNLPBaseDecoder,
    ):
        """

        :param encoder: Service's encoder.
        :param runner: Service's runner.
        :param decoder: Service's decoder.
        """
        self._encoder: ORKGNLPBaseEncoder = encoder
        self._runner: ORKGNLPBaseRunner = runner
        self._decoder: ORKGNLPBaseDecoder = decoder

    def run(self, raw_input: Any, **kwargs: Any) -> Any:
        """
        Executes a full pipline of the common service workflow:

        1. Runs the service encoder with the user's input.
        2. The encoded input is passed to the model runner, which in turn is executed.
        3. The model's output is decoded to a user-friendly format using the service's decoder.

        Note that the kwargs can be updated and passed through the pipeline components.

        :param raw_input: User's input to be encoded.
        :param kwargs: Named parameters for further processing config. Please check your used component documentation
            for specific parameter description.
        :return: The decoded user-friendly output.
        :raise orkgnlp.common.util.exceptions.ORKGNLPIllegalStateException: If either [Encoder, Runner, Decoder] is not
            initialized.
        """
        if not (self._encoder and self._runner and self._decoder):
            raise ORKGNLPIllegalStateError("Encoder, Runner and Decoder must be initialized!")

        inputs, additional_properties = self._encoder.encode(raw_input, **kwargs)
        kwargs.update(additional_properties)

        output, additional_properties = self._runner.run(inputs, **kwargs)
        kwargs.update(additional_properties)

        return self._decoder.decode(output, **kwargs)

    def release_memory(self):
        """
        Releases the memory of all available pipeline components.
        """
        self._encoder.release_memory()
        self._runner.release_memory()
        self._decoder.release_memory()


class ORKGNLPBaseConfig:
    """
    The ORKGNLPBaseConfig encapsulates the required configurations for a service given its name.
    """

    def __init__(self, service: str):
        """

        :param service: The service name.
        """
        service_data_dir = os.path.join(orkgnlp_context.get("ORKG_NLP_DATA_CACHE_ROOT"), service)

        self.service_name: str = service
        self.service_dir: str = service_data_dir
        self.requirements = self._get_requirement_paths()

    def _get_requirement_paths(self) -> Dict[str, str]:
        paths = {}
        for repo in orkgnlp_context.get("HUGGINGFACE_REPOS")[self.service_name]:
            for file_obj in repo["files"]:
                paths[file_obj["internal_name"]] = os.path.join(
                    self.service_dir, file_obj.get("subbdir", ""), file_obj["filename"]
                )

        return paths


class ORKGNLPBaseService:
    """
    Base class for shared config parameters and functionalities. All ORKG-NLP services must inherit from this class.

    This class follows the singleton pattern,  i.e. only one instance can be obtained from it or its subclasses.
    """

    @singleton
    def __new__(cls, *args, **kwargs):
        pass

    def __init__(
        self,
        service: str,
        *,
        force_download: bool = False,
        batch_size: int = 16,
        _unittest: bool = False
    ):
        """

        :param service: Service name.
        :param force_download: Indicates whether the required files are to be downloaded again. Defaults to False.
        :param batch_size: Size of the batches used during model prediction. This argument is used by services
            that applies batch evaluation. Defaults to 16.
        """
        self._pipeline_executors: Dict[str, PipelineExecutor] = {}
        self._config: ORKGNLPBaseConfig = ORKGNLPBaseConfig(service)
        self._force_download: bool = force_download
        self._batch_size: int = batch_size
        self._unittest: bool = _unittest
        self._download(service)

    def _download(self, service: str):
        """
        Downloads the required files for the given service name based on the ``force_download`` class attribute.

        :param service: a string representing a ORKG-NLP service name.
            Check :doc:`../services/services` for a full list.
        :return:
        """
        downloader.download(service, self._force_download)

    def _run(self, raw_input: Any, pipline_executor_name: str = None, **kwargs: Any):
        """
        Executes the only PipelineExecutor registered for the service, or one of them given its name.

        :param raw_input: User's input to be encoded.
        :param pipline_executor_name: Name of the PipelineExecutor to run.
        :param kwargs: Named parameters for further processing config. Please check your used component documentation
            for specific parameter description.
        :return: The decoded user-friendly output.
        :raise orkgnlp.common.util.exceptions.ORKGNLPIllegalStateException: If either [Encoder, Runner, Decoder] is not
            initialized.
        :raise orkgnlp.common.util.exceptions.ORKGNLPValidationError: If the given ``pipline_executor_name`` is unknown
            or not given, in case of multiple registered ones.
        """
        if not self._pipeline_executors:
            raise ORKGNLPIllegalStateError(
                "There is no PipelineRunner registered. Please consider registering"
                "one using the _register_pipline_runner() function."
            )

        if pipline_executor_name:
            if pipline_executor_name not in self._pipeline_executors:
                raise ORKGNLPValidationError("PipelineExecutor name is unknown.")

            return self._pipeline_executors[pipline_executor_name].run(raw_input, **kwargs)

        if len(self._pipeline_executors) > 1:
            raise ORKGNLPValidationError(
                "PipelineExecutor is ambiguous. Consider passing pipline_executor_name in the "
                "input."
            )

        return next(iter(self._pipeline_executors.values())).run(raw_input, **kwargs)

    def _register_pipeline(
        self,
        name: str,
        encoder: ORKGNLPBaseEncoder,
        runner: ORKGNLPBaseRunner,
        decoder: ORKGNLPBaseDecoder,
    ):
        """
        Registers a PipelineExecutor to the service.

        :param name: PipelineExecutors name.
        :param encoder: Service's encoder.
        :param runner: Service's runner.
        :param decoder: Service's decoder.
        """
        if name in self._pipeline_executors:
            raise ORKGNLPValidationError("PipelineExecutor name already exists.")

        self._pipeline_executors[name] = PipelineExecutor(encoder, runner, decoder)

    def release_memory(self):
        """
        Releases the memory of all available executors.
        """
        for executor in self._pipeline_executors.values():
            executor.release_memory()



# -*- coding: utf-8 -*-
""" BioAssays semantification service. """
from typing import Any

from orkgnlp.clustering.bioassays.decoder import BioassaysSemantifierDecoder
from orkgnlp.clustering.encoders import TfidfKmeansEncoder
from orkgnlp.common.config import orkgnlp_context
from orkgnlp.common.service.base import (
    ORKGNLPBaseDecoder,
    ORKGNLPBaseEncoder,
    ORKGNLPBaseRunner,
    ORKGNLPBaseService,
)
from orkgnlp.common.service.runners import ORKGNLPONNXRunner
from orkgnlp.common.util import io


class BioassaysSemantifier(ORKGNLPBaseService):
    """
    The BioassaysSemantifier  requires a clustering model, vectorizer and mapping.
    The required files are downloaded while initiation, if it has not happened before.

    You can pass the parameter ``force_download=True`` to remove and re-download the previous downloaded service files.
    """

    SERVICE_NAME = "bioassays-semantification"

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(self.SERVICE_NAME, *args, **kwargs)
        requirements = self._config.requirements

        encoder: ORKGNLPBaseEncoder = TfidfKmeansEncoder(io.read_onnx(requirements["vectorizer"]))
        runner: ORKGNLPBaseRunner = ORKGNLPONNXRunner(io.read_onnx(requirements["model"]))
        decoder: ORKGNLPBaseDecoder = BioassaysSemantifierDecoder(
            io.read_json(requirements["mapping"])
        )

        self._register_pipeline("main", encoder, runner, decoder)

    def __call__(self, text: str) -> any:
        """
        Semantifies a given BioAssay's description text.

        :param text: BioAssay's text to be semantified.
        :return: Dictionary object of semantified properties, resources and labels.
        """

        return self._run(raw_input=text)


orkgnlp_context.get("SERVICE_MAP")[BioassaysSemantifier.SERVICE_NAME] = BioassaysSemantifier



# -*- coding: utf-8 -*-
""" Common package. """

from orkgnlp.common import config, service, tools, util
from orkgnlp.common.service import base, runners

__all__ = ["config", "service", "tools", "util", "base", "runners"]



# -*- coding: utf-8 -*-
""" Defines decorators used within the package """
from typing import Type


def singleton(_):
    """
    This decorator can only be used above the __new__ function of a class. It's responsible for returning a pre-created
    instance of the respective class or create a new one, if not have happened before.

    :param _: The __new__ function.
    """

    def apply_pattern(cls: Type, *args, **kwargs):
        # attention: *args and **kwargs must be included even if not used!
        if not hasattr(cls, "instance"):
            cls.instance = super(cls.__class__, cls).__new__(cls)
        return cls.instance

    return apply_pattern


def sanitize_text(f):
    """
    This decorator can only be used above functions having the argument ``text`` at first position. The wrapper function
    returns the passed text if it's not empty or an empty string in case of empty or None passed argument.

    :param f: The decorated function.
    """

    def wrapper(text: str, *args):
        if not text:
            return f("", *args)

        return f(text, *args)

    return wrapper



# -*- coding: utf-8 -*-
""" BioAssays semantification service decoder. """

from typing import Any, Dict, Generator, Union

from overrides import overrides

from orkgnlp.common.service.base import ORKGNLPBaseDecoder


class BioassaysSemantifierDecoder(ORKGNLPBaseDecoder):
    """
    The BioassaysSemantifierDecoder decodes the Bioassays semantification service model's output
    to a user-friendly one.
    """

    def __init__(self, mapping: Dict[str, Dict[str, Any]]):
        """

        :param mapping: Cluster label resources mapping.
        :type mapping: Dict[str, Dict[str, Any]]
        """
        super().__init__()

        self._mapping = mapping

    @overrides
    def decode(self, model_output: Union[Any, Generator[Any, None, None]], **kwargs: Any) -> Any:
        cluster_label = model_output[0][0]
        return self._mapping[str(cluster_label)]["labels"]



# -*- coding: utf-8 -*-
""" Predicates recommendation service. """
from typing import Any

from orkgnlp.clustering.encoders import TransformerKmeansEncoder
from orkgnlp.clustering.predicates.decoder import PredicatesRecommenderDecoder
from orkgnlp.common.config import orkgnlp_context
from orkgnlp.common.service.base import (
    ORKGNLPBaseDecoder,
    ORKGNLPBaseEncoder,
    ORKGNLPBaseRunner,
    ORKGNLPBaseService,
)
from orkgnlp.common.service.runners import ORKGNLPONNXRunner
from orkgnlp.common.util import io


class PredicatesRecommender(ORKGNLPBaseService):
    """
    The PredicatesRecommender requires a clustering model, vectorizer, training set and predicates.
    The required files are downloaded while initiation, if it has not happened before.

    You can pass the parameter ``force_download=True`` to remove and re-download the previous downloaded service files.
    """

    SERVICE_NAME = "predicates-clustering"

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(self.SERVICE_NAME, *args, **kwargs)
        requirements = self._config.requirements

        encoder: ORKGNLPBaseEncoder = TransformerKmeansEncoder("allenai/scibert_scivocab_uncased")
        runner: ORKGNLPBaseRunner = ORKGNLPONNXRunner(io.read_onnx(requirements["model"]))
        decoder: ORKGNLPBaseDecoder = PredicatesRecommenderDecoder(
            io.read_df_from_json(requirements["training_data"], key="instances"),
            io.read_json(requirements["mapping"]),
        )
        self._register_pipeline("main", encoder, runner, decoder)

    def __call__(self, title: str, abstract: str):
        """
        Recommends predicates for a research paper.

        :param title: Title of the research paper.
        :param abstract: Abstract of the research paper.
        :return: List of predicates.
        """
        return self._run(
            raw_input="{} {}".format(title, abstract), output_names=["label", "labels_"]
        )


orkgnlp_context.get("SERVICE_MAP")[PredicatesRecommender.SERVICE_NAME] = PredicatesRecommender



# -*- coding: utf-8 -*-
""" Annotation services. """

from orkgnlp.annotation.agriner import AgriNer
from orkgnlp.annotation.csner import CSNer
from orkgnlp.annotation.tdm import TdmExtractor

__all__ = ["AgriNer", "CSNer", "TdmExtractor"]



# -*- coding: utf-8 -*-
""" CS-NER package. """
import contextlib
import io

import spacy

from orkgnlp.annotation.csner.annotator import CSNer

with contextlib.redirect_stdout(io.StringIO()):
    spacy.cli.download("en_core_web_md", False, False, "--quiet")

__all__ = ["CSNer"]



# -*- coding: utf-8 -*-
""" Agri-NER service decoder. """

from typing import Any, Dict, List

from overrides import overrides

from orkgnlp.common.service.base import ORKGNLPBaseDecoder


class AgriNerDecoder(ORKGNLPBaseDecoder):
    """
    The AgriNerDecoder decodes the Agri-NER service model's output
    to a user-friendly one.
    """

    CONCEPTS_MAP = {
        "RP": "RESEARCH_PROBLEM",
        "P": "PROCESS",
        "METH": "METHOD",
        "R": "RESOURCE",
        "S": "SOLUTION",
        "LOC": "LOCATION",
        "T": "TECHNOLOGY",
    }

    @overrides(check_signature=False)
    def decode(self, model_output: List[Dict[str, Any]], **kwargs: Any) -> Any:
        annotations = []
        seen = []
        for entity in model_output:
            if entity["entity_group"] in seen:
                continue

            seen.append(entity["entity_group"])
            annotations.append(
                {
                    "concept": self.CONCEPTS_MAP[entity["entity_group"]],
                    "entities": [
                        i["word"]
                        for i in model_output
                        if i["entity_group"] == entity["entity_group"]
                    ],
                }
            )

        return annotations



# -*- coding: utf-8 -*-
""" CS-NER service decoder. """
from typing import Any, Dict, Generator, List, Union

from overrides import overrides

from orkgnlp.common.service.base import ORKGNLPBaseDecoder


class CSNerDecoder(ORKGNLPBaseDecoder):
    """
    The CSNerDecoder decodes the CS-NER service model's output
    to a user-friendly one.
    """

    def __init__(self, alphabet):
        """

        :param alphabet: Dict representing the word, char and label alphabets.
        :type alphabet: Dict[str, Dict[str, int]]
        """
        super().__init__()

        self._alphabet = alphabet
        self._UNKNOWN = "</unk>"

    @overrides(check_signature=False)
    def decode(
        self,
        model_output: Union[Any, Generator[Any, None, None]],
        raw_texts,
        recover,
        **kwargs: Any
    ) -> Any:
        predicted_results = []

        for i, batch in enumerate(model_output):
            _, nbest_tag_seq = batch
            tag_seq = nbest_tag_seq[:, :, 0]
            pred_label, _ = self._recover_label(
                tag_seq, recover[i][0], recover[i][0], recover[i][2]
            )
            predicted_results += pred_label

        entities = self._get_entities(raw_texts, predicted_results)
        return self._prepare_annotations(entities)

    def _recover_label(self, pred_variable, gold_variable, mask_variable, word_recover):
        """
        input:
            pred_variable (batch_size, sent_len): pred tag result
            gold_variable (batch_size, sent_len): gold result variable
            mask_variable (batch_size, sent_len): mask variable
        """
        pred_variable = pred_variable[word_recover]
        gold_variable = gold_variable[word_recover]
        mask_variable = mask_variable[word_recover]
        batch_size = gold_variable.size(0)
        seq_len = gold_variable.size(1)
        mask = mask_variable.cpu().data.numpy()
        pred_tag = pred_variable.cpu().data.numpy()
        gold_tag = gold_variable.cpu().data.numpy()
        batch_size = mask.shape[0]
        pred_label = []
        gold_label = []

        for idx in range(batch_size):
            labels = list(self._alphabet["label"].keys())

            pred = [
                self._get_instance(labels, pred_tag[idx][idy])
                for idy in range(seq_len)
                if mask[idx][idy] != 0
            ]
            gold = [
                self._get_instance(labels, gold_tag[idx][idy])
                for idy in range(seq_len)
                if mask[idx][idy] != 0
            ]

            pred_label.append(pred)
            gold_label.append(gold)

        return pred_label, gold_label

    @staticmethod
    def _get_entities(raw_texts, predicted_results):
        entities = {}

        for idx in range(len(predicted_results)):
            sent_length = len(predicted_results[idx])
            key = ""
            entity_txt = ""

            for idy in range(sent_length):
                if "S-" in predicted_results[idx][idy]:
                    key = predicted_results[idx][idy].split("-")[1]

                    if key in entities.keys():
                        entities[key].append(raw_texts[idx][0][idy])
                    else:
                        entities[key] = [raw_texts[idx][0][idy]]

                elif "B-" in predicted_results[idx][idy]:
                    key = predicted_results[idx][idy].split("-")[1]
                    entity_txt = raw_texts[idx][0][idy].strip()

                elif "I-" in predicted_results[idx][idy]:
                    entity_txt += " " + raw_texts[idx][0][idy].strip()

                elif "E-" in predicted_results[idx][idy]:
                    entity_txt += " " + raw_texts[idx][0][idy].strip()

                    if key in entities.keys():
                        entities[key].append(entity_txt.strip())
                    else:
                        entities[key] = [entity_txt.strip()]

                    key = ""
                    entity_txt = ""

        return entities

    @staticmethod
    def _get_instance(instances, index):
        if index == 0:
            return instances[0]
        try:
            return instances[index - 1]
        except IndexError:
            print("WARNING:Alphabet get_instance, unknown instance, return the first label.")
            return instances[0]

    @staticmethod
    def _prepare_annotations(entities: Dict[str, List[Any]]) -> List[Dict[str, Any]]:
        annotations = []

        for concept in entities:
            annotations.append({"concept": concept, "entities": entities[concept]})

        return annotations



# -*- coding: utf-8 -*-
""" Templates Recommendation service. """
from typing import Any

from transformers import BertForSequenceClassification

from orkgnlp.common.config import orkgnlp_context
from orkgnlp.common.service.base import (
    ORKGNLPBaseDecoder,
    ORKGNLPBaseEncoder,
    ORKGNLPBaseRunner,
    ORKGNLPBaseService,
)
from orkgnlp.common.service.runners import ORKGNLPTorchRunner
from orkgnlp.common.util import io
from orkgnlp.nli.templates.decoder import TemplatesRecommenderDecoder
from orkgnlp.nli.templates.encoder import TemplatesRecommenderEncoder


class TemplatesRecommender(ORKGNLPBaseService):
    """
    The TemplatesRecommender requires a transformers.BertForSequenceClassification pretrained model and the templates
    used for training as premises.
    The required files are downloaded while initiation, if it has not happened before.

    You can pass the parameter ``force_download=True`` to remove and re-download the previous downloaded service files.
    """

    SERVICE_NAME = "templates-recommendation"

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(self.SERVICE_NAME, *args, **kwargs)
        templates = io.read_json(self._config.requirements["labels"])["templates"]

        encoder: ORKGNLPBaseEncoder = TemplatesRecommenderEncoder(templates)
        runner: ORKGNLPBaseRunner = ORKGNLPTorchRunner(
            io.load_transformers_pretrained(
                self._config.service_dir, BertForSequenceClassification
            )
        )
        decoder: ORKGNLPBaseDecoder = TemplatesRecommenderDecoder(templates)

        self._register_pipeline("main", encoder, runner, decoder)

    def __call__(self, title: str, abstract: str, top_n: int = 5):
        """
        Recommends templates for a research paper.

        :param title: Title of the research paper.
        :param abstract: Abstract of the research paper.
        :param top_n: Top n results to be extracted. Defaults to 5.
        :return: List of templates.
        """
        return self._run(
            raw_input="{} {}".format(title, abstract),
            top_n=top_n,
            multiple_batches=True,
        )


orkgnlp_context.get("SERVICE_MAP")[TemplatesRecommender.SERVICE_NAME] = TemplatesRecommender



# -*- coding: utf-8 -*-
"""
Root package of orkgnlp.
"""

__version__ = "0.10.0"

import logging

from orkgnlp import annotation, clustering, nli
from orkgnlp.common import config, service, tools, util
from orkgnlp.common.tools import download, load

__all__ = [
    "annotation",
    "clustering",
    "nli",
    "config",
    "service",
    "tools",
    "util",
    "download",
    "load",
]


# Root logger configuration
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

stdout = logging.StreamHandler()
stdout.setLevel(logging.DEBUG)

formatter = logging.Formatter("%(name)s - %(levelname)s: %(message)s")
stdout.setFormatter(formatter)

logger.addHandler(stdout)



# -*- coding: utf-8 -*-
""" Common encoders for the clustering services. """
from typing import Any, Dict, Tuple

from onnx import ModelProto
from overrides import overrides
from sentence_transformers import SentenceTransformer

from orkgnlp.common.service.base import ORKGNLPBaseEncoder
from orkgnlp.common.service.runners import ORKGNLPONNXRunner
from orkgnlp.common.util import text


class TfidfKmeansEncoder(ORKGNLPBaseEncoder):
    """
    The TfidfKmeansEncoder encodes the given input to a TF-IDF vector
    needed to execute a Kmeans onnx model.
    """

    def __init__(self, vectorizer: ModelProto):
        """

        :param vectorizer: The TF-IDF vectorizer needed for the encoding.
        """
        super().__init__()
        self._vectorizer: ORKGNLPONNXRunner = ORKGNLPONNXRunner(vectorizer)

    @overrides
    def encode(self, raw_input: Any, **kwargs: Any) -> Tuple[Any, Dict[str, Any]]:
        preprocessed_text = self._text_process(raw_input)
        output, _ = self._vectorizer.run(inputs=([preprocessed_text],), output_names=["variable"])
        return (output[0][0],), kwargs

    @staticmethod
    def _text_process(q: str) -> str:
        q = text.remove_punctuation(q)
        q = text.remove_stopwords(q)
        q = text.lemmatize(q)
        return q.lower()


class TransformerKmeansEncoder(ORKGNLPBaseEncoder):
    """
    The SciBERTKmeansEncoder encodes the given input to a SciBERT vector
    needed to execute a Kmeans onnx model.
    """

    def __init__(self, transformer_path: str):
        """

        :param transformer_path: Path to transformers model. Can be a model name on Huggingface.
        """
        super().__init__()
        self._vectorizer: SentenceTransformer = SentenceTransformer(transformer_path)

    @overrides
    def encode(self, raw_input: Any, **kwargs: Any) -> Tuple[Any, Dict[str, Any]]:
        return (self._vectorizer.encode([raw_input])[0],), kwargs



# -*- coding: utf-8 -*-
""" Utility file for text processing functionalities. """
import re
import string
from typing import List, Union

import nltk
from nltk.corpus import stopwords
from nltk.stem import WordNetLemmatizer

from orkgnlp.common.util.decorators import sanitize_text


@sanitize_text
def remove_punctuation(text: str) -> str:
    """
    Removes punctuations from the given ``text``.

    :param text:
    :return: text with punctuations removed.
    """
    return text.translate(str.maketrans("", "", string.punctuation))


@sanitize_text
def remove_stopwords(text: str, language: str = "english") -> str:
    """
    Removes stopwords from the given ``text``.
    Uses the stopwords defined in ``NLTK``.

    :param text:
    :param language: defaults to ``english``.
    :return: text with stopwords removed.
    """
    nltk.download("stopwords", quiet=True)

    return " ".join([word for word in text.split() if word not in stopwords.words(language)])


@sanitize_text
def lemmatize(text: str) -> str:
    """
    Lemmatizes the words in the given ``text``.
    Uses the WordNetLemmatizer defined in ``NLTK``.

    :param text:
    :return: text with its lemmatized words.
    """
    nltk.download("wordnet", quiet=True)

    stemmer = WordNetLemmatizer()
    return " ".join([stemmer.lemmatize(word) for word in text.split()])


@sanitize_text
def replace(text: str, chars: Union[str, List[str]], replacement: str) -> str:
    """
    Replaces the occurrences of any item in ``chars`` in ``text`` with ``replacement``.

    :param text:
    :param chars: Regex or list of regexes that will be separated with | as OR operator,
        their occurrences to be replaced.
    :param replacement: The replacement string
    """
    chars = chars if isinstance(chars, list) else [chars]
    regex = "|".join(chars)
    return re.sub(regex, replacement, text)


@sanitize_text
def trim(text: str) -> str:
    """
    Removes any whitespace character and returns the same input only with single spaces.

    :param text:
    """
    return " ".join(text.split())



# -*- coding: utf-8 -*-
""" Agri-NER service. """

from typing import Any

from transformers import pipeline

from orkgnlp.annotation.agriner.decoder import AgriNerDecoder
from orkgnlp.common.config import orkgnlp_context
from orkgnlp.common.service.base import (
    ORKGNLPBaseDecoder,
    ORKGNLPBaseEncoder,
    ORKGNLPBaseRunner,
    ORKGNLPBaseService,
)
from orkgnlp.common.service.runners import ORKGNLPTorchRunner


class AgriNer(ORKGNLPBaseService):
    """
    The AgriNer requires a classification model trained on titles and its configurations obtained during the training.
    The required files are downloaded while initiation, if it has not happened before.

    You can pass the parameter ``force_download=True`` to remove and re-download the previous downloaded service files.
    """

    SERVICE_NAME = "agri-ner"

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(self.SERVICE_NAME, *args, **kwargs)

        encoder: ORKGNLPBaseEncoder = ORKGNLPBaseEncoder()
        _model = pipeline(
            task="token-classification",
            model=self._config.service_dir,
            tokenizer="bert-base-cased",
            aggregation_strategy="simple",
        )
        runner: ORKGNLPBaseRunner = ORKGNLPTorchRunner(_model)
        decoder: ORKGNLPBaseDecoder = AgriNerDecoder()
        self._register_pipeline("main", encoder, runner, decoder)

    def __call__(self, title: str) -> Any:
        """
        Applies Named Entity Recognition on the given paper's title.

        :param title: Paper's title.
        :return: A list of the annotated parts for the given text is returned.
        """
        return self._run(raw_input=title)


orkgnlp_context.get("SERVICE_MAP")[AgriNer.SERVICE_NAME] = AgriNer



# -*- coding: utf-8 -*-
""" CS-NER service encoder. """

import sys
from typing import Any, Dict, Tuple

import numpy as np
import spacy
import torch
from overrides import overrides

from orkgnlp.common.service.base import ORKGNLPBaseEncoder


class CSNerEncoder(ORKGNLPBaseEncoder):
    """
    The CSNerEncoder encodes the given input to the arguments
    needed to execute a Sequence labeling TorchScript model.
    """

    def __init__(self, alphabet: Dict[str, Dict[str, int]]):
        """

        :param alphabet: Dict representing the word, char and label alphabets.
        :type alphabet: Dict[str, Dict[str, int]]
        """
        super().__init__()

        self._alphabet = alphabet
        self._UNKNOWN = "</unk>"
        self._spacy_nlp = spacy.load(
            "en_core_web_md",
            disable=["tokenizer", "tagger", "ner", "textcat", "lemmatizer"],
        )

    @overrides
    def encode(self, raw_input: Any, **kwargs: Any) -> Tuple[Any, Dict[str, Any]]:
        raw_texts, raw_ids = self._read_instance(raw_input)

        batch_size = 10
        total_batch = len(raw_ids) // batch_size + 1
        model_input = []
        decoder_input = {"raw_texts": raw_texts, "recover": []}

        for batch_id in range(total_batch):
            start = batch_id * batch_size
            end = (batch_id + 1) * batch_size

            if end > len(raw_ids):
                end = len(raw_ids)

            instance = raw_ids[start:end]
            if not instance:
                continue

            (
                word,
                features,
                word_len,
                word_recover,
                char,
                char_len,
                char_recover,
                label,
                mask,
            ) = self._batchify(instance)
            model_input.append(
                (
                    word,
                    features,
                    word_len,
                    char,
                    char_recover,
                    mask,
                    1,
                )
            )
            decoder_input["recover"].append((label, mask, word_recover))

        kwargs.update(decoder_input)
        return model_input, kwargs

    def _read_instance(self, q, char_padding_size=-1, char_padding_symbol="</pad>"):
        document = self._spacy_nlp(q)
        instance_texts = []
        instance_ids = []

        # for sequence labeling text format
        for span in document.sents:
            (
                words,
                features,
                chars,
                labels,
                word_ids,
                feature_ids,
                char_ids,
                label_ids,
            ) = ([], [], [], [], [], [], [], [])
            sentence = [document[i] for i in range(span.start, span.end)]

            for token in sentence:
                word = token.text.strip()

                if not word:
                    continue

                if sys.version_info[0] < 3:
                    word = word.decode("utf-8")

                words.append(word)

                if self._alphabet["number_normalized"]:
                    word = self._normalize_word(word)

                label = "O"
                labels.append(label)
                word_ids.append(
                    self._alphabet["word"].get(word) or self._alphabet["word"].get(self._UNKNOWN)
                )
                label_ids.append(
                    self._alphabet["label"].get(label)
                    or self._alphabet["label"].get(self._UNKNOWN)
                )

                # get features
                feat_list = []
                feat_Id = []

                features.append(feat_list)
                feature_ids.append(feat_Id)

                # get char
                char_list = []
                char_id = []
                for char in word:
                    char_list.append(char)

                if char_padding_size > 0:
                    char_number = len(char_list)
                    if char_number < char_padding_size:
                        char_list = char_list + [char_padding_symbol] * (
                            char_padding_size - char_number
                        )
                    assert len(char_list) == char_padding_size
                else:
                    # not padding
                    pass

                for char in char_list:
                    char_id.append(
                        self._alphabet["char"].get(char)
                        or self._alphabet["char"].get(self._UNKNOWN)
                    )

                chars.append(char_list)
                char_ids.append(char_id)

            instance_texts.append([words, features, chars, labels])
            instance_ids.append([word_ids, feature_ids, char_ids, label_ids])

        return instance_texts, instance_ids

    @staticmethod
    def _batchify(input_batch_list, if_train=False):
        """
        input: list of words, chars and labels, various length.
            [[words, features, chars, labels],[words, features, chars,labels],...]
            words: word ids for one sentence. (batch_size, sent_len)
            features: features ids for one sentence. (batch_size, sent_len, feature_num)
            chars: char ids for on sentences, various length. (batch_size, sent_len, each_word_length)
            labels: label ids for one sentence. (batch_size, sent_len)

        output:
            zero padding for word and char, with their batch length
            word_seq_tensor: (batch_size, max_sent_len) Variable
            feature_seq_tensors: [(batch_size, max_sent_len),...] list of Variable
            word_seq_lengths: (batch_size,1) Tensor
            char_seq_tensor: (batch_size*max_sent_len, max_word_len) Variable
            char_seq_lengths: (batch_size*max_sent_len,1) Tensor
            char_seq_recover: (batch_size*max_sent_len,1)  recover char sequence order
            label_seq_tensor: (batch_size, max_sent_len)
            mask: (batch_size, max_sent_len)
        """
        batch_size = len(input_batch_list)
        words = [sent[0] for sent in input_batch_list]
        features = [np.asarray(sent[1]) for sent in input_batch_list]
        feature_num = len(features[0][0])
        chars = [sent[2] for sent in input_batch_list]
        labels = [sent[3] for sent in input_batch_list]
        word_seq_lengths = torch.LongTensor(list(map(len, words)))
        max_seq_len = word_seq_lengths.max().item()
        word_seq_tensor = torch.zeros((batch_size, max_seq_len), requires_grad=if_train).long()
        label_seq_tensor = torch.zeros((batch_size, max_seq_len), requires_grad=if_train).long()
        feature_seq_tensors = []
        for idx in range(feature_num):
            feature_seq_tensors.append(
                torch.zeros((batch_size, max_seq_len), requires_grad=if_train).long()
            )
        mask = torch.zeros((batch_size, max_seq_len), requires_grad=if_train).bool()
        for idx, (seq, label, seqlen) in enumerate(zip(words, labels, word_seq_lengths)):
            seqlen = seqlen.item()
            word_seq_tensor[idx, :seqlen] = torch.LongTensor(seq)
            label_seq_tensor[idx, :seqlen] = torch.LongTensor(label)
            mask[idx, :seqlen] = torch.Tensor([1] * seqlen)
            for idy in range(feature_num):
                feature_seq_tensors[idy][idx, :seqlen] = torch.LongTensor(features[idx][:, idy])
        word_seq_lengths, word_perm_idx = word_seq_lengths.sort(0, descending=True)
        word_seq_tensor = word_seq_tensor[word_perm_idx]
        for idx in range(feature_num):
            feature_seq_tensors[idx] = feature_seq_tensors[idx][word_perm_idx]

        label_seq_tensor = label_seq_tensor[word_perm_idx]
        mask = mask[word_perm_idx]
        # deal with char
        # pad_chars (batch_size, max_seq_len)
        pad_chars = [
            chars[idx] + [[0]] * (max_seq_len - len(chars[idx])) for idx in range(len(chars))
        ]
        length_list = [list(map(len, pad_char)) for pad_char in pad_chars]
        max_word_len = max(map(max, length_list))
        char_seq_tensor = torch.zeros(
            (batch_size, max_seq_len, max_word_len), requires_grad=if_train
        ).long()
        char_seq_lengths = torch.LongTensor(length_list)
        for idx, (seq, seqlen) in enumerate(zip(pad_chars, char_seq_lengths)):
            for idy, (word, wordlen) in enumerate(zip(seq, seqlen)):
                # print len(word), wordlen
                char_seq_tensor[idx, idy, :wordlen] = torch.LongTensor(word)

        char_seq_tensor = char_seq_tensor[word_perm_idx].view(batch_size * max_seq_len, -1)
        char_seq_lengths = char_seq_lengths[word_perm_idx].view(
            batch_size * max_seq_len,
        )
        char_seq_lengths, char_perm_idx = char_seq_lengths.sort(0, descending=True)
        char_seq_tensor = char_seq_tensor[char_perm_idx]
        _, char_seq_recover = char_perm_idx.sort(0, descending=False)
        _, word_seq_recover = word_perm_idx.sort(0, descending=False)
        return (
            word_seq_tensor,
            torch.tensor(feature_seq_tensors),
            word_seq_lengths,
            word_seq_recover,
            char_seq_tensor,
            char_seq_lengths,
            char_seq_recover,
            label_seq_tensor,
            mask,
        )

    @staticmethod
    def _normalize_word(word):
        new_word = ""

        for char in word:
            if char.isdigit():
                new_word += "0"
            else:
                new_word += char

        return new_word



# -*- coding: utf-8 -*-
"""
Includes Input/Output (I/O) functionalities like reading and writing from and into specific file formats.
"""

import json
import pickle
from typing import Any, Dict, Type

import onnx
import pandas as pd
import torch
from onnx import ModelProto
from pandas import DataFrame
from torch.jit import ScriptModule
from transformers import PreTrainedModel


def read_json(input_path: str) -> Dict[str, Any]:
    """
    Reads the ``json`` file of the given ``input_path``.

    :param input_path: Path to the json file
    :return: A loaded json object.
    """
    with open(input_path, encoding="utf-8") as f:
        json_data = json.load(f)

    return json_data


def read_df_from_json(input_path: str, key: str = None) -> DataFrame:
    """
    Reads the ``json`` file of the given ``input_path`` and converts it to pandas ``dataframe``.

    :param input_path: Path to the json file
    :param key: Specifies the object to be converted.
    :return: A loaded pandas dataframe object.
    """
    json_file = read_json(input_path)

    if key:
        return pd.json_normalize(json_file[key])

    return pd.json_normalize(json_file)


def read_onnx(input_path: str) -> ModelProto:
    """
    Reads the ``onnx`` file of the given ``input_path``.

    :param input_path: Path to the onnx file
    :return: A loaded onnx object.
    """
    return onnx.load(input_path)


def read_pickle(input_path: str) -> Any:
    """
    Reads the ``pickle`` file of the given ``input_path``.

    :param input_path: Path to the pickle file
    :return: A loaded pickle object.
    """
    with open(input_path, "rb") as f:
        loaded_object = pickle.load(f)

    return loaded_object


def read_csv(input_path: str, header: str = None, sep: str = ",") -> DataFrame:
    """
    Reads the ``csv`` file of the given ``input_path``.

    :param input_path: Path to the csv file.
    :param header: See pandas.read_csv. Defaults to None
    :param sep: See pandas.read_csv. Defaults to ','
    :return: str
    """
    return pd.read_csv(input_path, header=header, sep=sep)


def load_torch_jit(input_path: str, device_name: str = "cpu") -> ScriptModule:
    """
    Loads the scripted/traced ``torch model`` file (ScriptModule) of the given ``input_path``.

    :param input_path: Path to the scripted/traced torch model file (ScriptModule).
    :param device_name: Defaults to ``cpu``
    :return: A loaded torch model (ScriptModule) object.
    """

    return torch.jit.load(input_path, map_location=torch.device(device_name))


def load_transformers_pretrained(
    input_path: str, transformers_cls: Type[PreTrainedModel]
) -> PreTrainedModel:
    """
    Loads a transformers model weights given of the given ``input_path`` to the given ``transformers_cls``.

    :param input_path: Path to the pretrained transformers model folder
    :param transformers_cls: The class type of the pretrained model you need to load.
    :return:
    """

    return transformers_cls.from_pretrained(input_path)



# -*- coding: utf-8 -*-
""" TDM-Extraction service encoder. """

from typing import Any, Dict, List, Tuple

import torch
from overrides import overrides
from pandas import DataFrame
from torch.utils.data import DataLoader, Dataset
from transformers import PreTrainedTokenizer, XLNetTokenizer

from orkgnlp.common.service.base import ORKGNLPBaseEncoder


class TdmExtractorEncoder(ORKGNLPBaseEncoder):
    """
    The TdmExtractorEncoder encodes the given input to the arguments
    needed to execute an XLNetForSequenceClassification model.
    """

    def __init__(self, labels: DataFrame, batch_size: int):
        """

        :param labels: TDM gold labels given as one-columned-dataframe
        :param batch_size: Size of the batches used during model prediction.
        """
        super().__init__()

        self.labels: DataFrame = labels
        self.batch_size: int = batch_size
        self.tokenizer: XLNetTokenizer = XLNetTokenizer.from_pretrained("xlnet-base-cased")
        self.max_input_sizes: int = self.tokenizer.max_model_input_sizes["xlnet-base-cased"] or 512
        self.device: str = "cpu"

    @overrides
    def encode(self, raw_input: Any, **kwargs: Any) -> Tuple[Any, Dict[str, Any]]:
        dataset = TdmDataset(raw_input, self.labels, self.tokenizer, self.max_input_sizes)
        dataloader = DataLoader(
            dataset,
            batch_size=self.batch_size,
            shuffle=False,
            collate_fn=self._collate_batch,
        )

        def batch_generator():
            for batch in dataloader:
                yield {
                    "input_ids": batch[0].to(self.device),
                    "token_type_ids": batch[1].to(self.device),
                    "attention_mask": batch[2].to(self.device),
                }

        return batch_generator(), kwargs

    @staticmethod
    def _collate_batch(batch: List[Tuple[Any, ...]]) -> Tuple[Any, ...]:
        max_shape = max([instance[0].shape[0] for instance in batch])

        args_num = len(batch[0])
        args = [[] for _ in range(args_num)]
        for instance in batch:
            for i in range(args_num):
                zero_padding = torch.zeros(max_shape - instance[i].shape[0], dtype=torch.int)
                args[i].append(torch.cat((instance[i], zero_padding)))

        return tuple(torch.stack(args[i]) for i in range(args_num))


class TdmDataset(Dataset):
    """
    The TdmDataset is a customized torch.utils.data.Dataset that simplifies the tokenization of sequences and
    can be used afterwards in a torch.utils.data.Dataloader for batch creation.
    """

    def __init__(
        self,
        text: str,
        labels: DataFrame,
        tokenizer: PreTrainedTokenizer,
        max_input_sizes: int,
    ):
        """

        :param text: Input text (hypothesis) to be concatenated with all known labels (premises).
        :param labels: TDM gold labels given as one-columned-dataframe
        :param tokenizer: Tokenizer for tokenizing the texts.
        :param max_input_sizes: Max length of a sequence including special characters.
        """
        self.labels: DataFrame = labels
        self.tokenizer: PreTrainedTokenizer = tokenizer
        self.hypothesis_ids: List[int] = self.tokenizer.encode(text, add_special_tokens=False)
        self.max_input_sizes: int = max_input_sizes

    def __len__(self):
        return len(self.labels.index)

    def __getitem__(self, idx):
        premise_ids = self.tokenizer.encode(
            self.labels.iloc[idx].tolist()[0], add_special_tokens=False
        )

        # -3 to account for the special characters
        self._truncate_seq_pair(premise_ids, self.hypothesis_ids, self.max_input_sizes - 3)

        sequence_token_ids = [
            self.tokenizer.cls_token_id,
            *premise_ids,
            self.tokenizer.sep_token_id,
            *self.hypothesis_ids,
            self.tokenizer.sep_token_id,
        ]
        token_type_ids = self._get_token_type(sequence_token_ids, self.tokenizer.sep_token_id)
        attention_mask_ids = [1] * len(sequence_token_ids)

        return (
            torch.tensor(sequence_token_ids),
            torch.tensor(token_type_ids),
            torch.tensor(attention_mask_ids),
        )

    @staticmethod
    def _get_token_type(tokens: List[int], sep: int) -> List[int]:
        sep_index = tokens.index(sep) + 1
        return [0] * sep_index + [1] * (len(tokens) - sep_index)

    @staticmethod
    def _truncate_seq_pair(tokens_a: List[int], tokens_b: List[int], max_length: int):
        """Truncates a sequence pair in place to the maximum length."""

        # This is a simple heuristic which will always truncate the longer sequence
        # one token at a time. This makes more sense than truncating an equal percent
        # of tokens from each, since if one sequence is very short then each token
        # that's truncated likely contains more information than a longer sequence.
        while True:
            total_length = len(tokens_a) + len(tokens_b)
            if total_length <= max_length:
                break
            if len(tokens_a) > len(tokens_b):
                tokens_a.pop()
            else:
                tokens_b.pop()



# -*- coding: utf-8 -*-
"""
Tooling package that provides a set of different tools.
"""
from orkgnlp.common.tools.downloader import download
from orkgnlp.common.tools.loader import load

__all__ = ["download", "load"]



# -*- coding: utf-8 -*-
"""
Clustering services.
"""

from orkgnlp.clustering.bioassays import BioassaysSemantifier
from orkgnlp.clustering.predicates import PredicatesRecommender

__all__ = ["BioassaysSemantifier", "PredicatesRecommender"]



# -*- coding: utf-8 -*-
from typing import Any

from orkgnlp.common.config import orkgnlp_context
from orkgnlp.common.service.base import ORKGNLPBaseService
from orkgnlp.common.util.exceptions import ORKGNLPUnknownServiceError


def load(service_name: str, *args: Any, **kwargs: Any) -> ORKGNLPBaseService:
    """
    Creates a class object of the requested service.

    :param service_name: The name of the service class.
    :return: A service class object.
    """
    try:
        return orkgnlp_context["SERVICE_MAP"][service_name](*args, **kwargs)
    except KeyError:
        raise ORKGNLPUnknownServiceError(
            "{} is not known.  Please use one of: {}".format(
                service_name, list(orkgnlp_context["SERVICE_MAP"].keys())
            )
        )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['orkgnlp',
 'orkgnlp.annotation',
 'orkgnlp.annotation.agriner',
 'orkgnlp.annotation.csner',
 'orkgnlp.annotation.tdm',
 'orkgnlp.clustering',
 'orkgnlp.clustering.bioassays',
 'orkgnlp.clustering.predicates',
 'orkgnlp.common',
 'orkgnlp.common.config',
 'orkgnlp.common.service',
 'orkgnlp.common.tools',
 'orkgnlp.common.util',
 'orkgnlp.nli',
 'orkgnlp.nli.templates']

package_data = \
{'': ['*']}

install_requires = \
['huggingface-hub>=0.5.1,<0.6.0',
 'nltk==3.5',
 'numpy==1.21.6',
 'onnx==1.11.0',
 'onnxruntime==1.11.1',
 'overrides>=6.1.0,<7.0.0',
 'pandas==1.3.5',
 'pre-commit>=3.1.0,<4.0.0',
 'protobuf==3.20.0',
 'sentence-transformers==2.2.2',
 'sentencepiece>=0.1.96,<0.2.0',
 'spacy==3.3.0',
 'torch==1.11.0',
 'transformers==4.19.3',
 'typing-extensions>=3.7.4,<4.0.0']

entry_points = \
{'console_scripts': ['test = tests.unittests:main']}

setup_kwargs = {
    'name': 'orkgnlp',
    'version': '0.10.0',
    'description': 'Python package wrapping the ORKG NLP Services.',
    'long_description': '# ORKG NLP PyPI\n[![Documentation Status](https://readthedocs.org/projects/orkg-nlp-pypi/badge/?version=latest)](https://orkg-nlp-pypi.readthedocs.io/en/latest/?badge=latest)\n[![PyPI version](https://badge.fury.io/py/orkgnlp.svg)](https://badge.fury.io/py/orkgnlp)\n[![pipeline status](https://gitlab.com/TIBHannover/orkg/nlp/orkg-nlp-pypi/badges/main/pipeline.svg)](https://gitlab.com/TIBHannover/orkg/nlp/orkg-nlp-pypi/-/commits/main)\n[![pre-commit](https://img.shields.io/badge/pre--commit-enabled-brightgreen?logo=pre-commit)](https://github.com/pre-commit/pre-commit)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n[![Imports: isort](https://img.shields.io/badge/%20imports-isort-%231674b1?style=flat&labelColor=ef8336)](https://pycqa.github.io/isort/)\n\nPyPI package wrapping the ORKG NLP services.\n\nCheck our [Read the Docs](https://orkg-nlp-pypi.readthedocs.io/en/latest/) for more details!\n\nYour [contribution](https://orkg-nlp-pypi.readthedocs.io/en/latest/contribute.html) to `orkgnlp` will always be honored!\n',
    'author': 'Omar Arab Oghli',
    'author_email': 'omar.araboghli@outlook.com',
    'maintainer': 'Omar Arab Oghli',
    'maintainer_email': 'None',
    'url': 'http://orkg.org/about',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<3.10',
}


setup(**setup_kwargs)



# -*- coding: utf-8 -*-
"""
Config package that allows a global configuring of orkg-nlp.
"""

from orkgnlp.common.config.context import (
    orkgnlp_context,
    set_data_cache_root,
    set_verbosity,
)

__all__ = ["orkgnlp_context", "set_verbosity", "set_data_cache_root"]



# -*- coding: utf-8 -*-
""" Templates Recommendation service encoder. """

from typing import Any, Dict, List, Tuple

import torch
from overrides import overrides
from transformers import BertTokenizer

from orkgnlp.common.service.base import ORKGNLPBaseEncoder
from orkgnlp.common.util import text


class TemplatesRecommenderEncoder(ORKGNLPBaseEncoder):
    """
    The TemplatesRecommenderEncoder encodes the given input to the arguments
    needed to execute a BertForSequenceClassification model.
    """

    def __init__(self, templates: List[Dict[str, str]]):
        """

        :param templates: templates used for training the service models as premises.
        """
        super().__init__()

        self.templates: List[Dict[str, str]] = templates
        self.tokenizer: BertTokenizer = BertTokenizer.from_pretrained(
            "allenai/scibert_scivocab_uncased"
        )
        self.max_input_sizes: int = (
            self.tokenizer.max_model_input_sizes["bert-base-uncased"] or 512
        )
        self.device: str = "cpu"

    @overrides
    def encode(self, raw_input: Any, **kwargs: Any) -> Tuple[Any, Dict[str, Any]]:
        def batch_generator():
            for template in self.templates:
                sequence = "[CLS] {} [SEP] {} [SEP]".format(
                    self._post_process(template["premise"]),
                    self._post_process(raw_input),
                )
                sequence_tokens = self.tokenizer.tokenize(sequence)

                max_sizes = self.max_input_sizes
                attention_mask = self._get_attention_mask(sequence_tokens)[:max_sizes]
                token_type = self._get_token_type(sequence_tokens)[:max_sizes:]
                sequence_tokens = self.tokenizer.convert_tokens_to_ids(sequence_tokens)[
                    :max_sizes:
                ]

                attention_mask = torch.tensor(attention_mask).unsqueeze(0).to(self.device)
                token_type = torch.tensor(token_type).unsqueeze(0).to(self.device)
                sequence_tokens = torch.tensor(sequence_tokens).unsqueeze(0).to(self.device)

                yield {
                    "input_ids": sequence_tokens,
                    "token_type_ids": token_type,
                    "attention_mask": attention_mask,
                }

        return batch_generator(), kwargs

    @staticmethod
    def _get_attention_mask(tokens):
        return [1] * len(tokens)

    @staticmethod
    def _get_token_type(tokens):
        sep_index = tokens.index("[SEP]") + 1
        return [0] * sep_index + [1] * (len(tokens) - sep_index)

    @staticmethod
    def _post_process(string):
        string = text.replace(string, [r"\s+-\s+", "-", "_", r"\."], " ")
        return text.trim(string).lower()



# -*- coding: utf-8 -*-
""" TDM-Extraction service. """
from typing import Any

from pandas import DataFrame
from transformers import XLNetForSequenceClassification

from orkgnlp.annotation.tdm.decoder import TdmExtractorDecoder
from orkgnlp.annotation.tdm.encoder import TdmExtractorEncoder
from orkgnlp.common.config import orkgnlp_context
from orkgnlp.common.service.base import (
    ORKGNLPBaseDecoder,
    ORKGNLPBaseEncoder,
    ORKGNLPBaseRunner,
    ORKGNLPBaseService,
)
from orkgnlp.common.service.runners import ORKGNLPTorchRunner
from orkgnlp.common.util import io


class TdmExtractor(ORKGNLPBaseService):
    """
    The TdmExtractor requires a transformers.XLNetForSequenceClassification pretrained model and a TDM gold labels file.
    The required files are downloaded while initiation, if it has not happened before.

    You can pass the parameter ``force_download=True`` to remove and re-download the previous downloaded service files.
    """

    SERVICE_NAME = "tdm-extraction"

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(self.SERVICE_NAME, *args, **kwargs)
        requirements = self._config.requirements

        labels: DataFrame = io.read_csv(requirements["labels"], sep="\t")

        if self._unittest:
            labels = labels.head()

        encoder: ORKGNLPBaseEncoder = TdmExtractorEncoder(labels, self._batch_size)
        runner: ORKGNLPBaseRunner = ORKGNLPTorchRunner(
            io.load_transformers_pretrained(
                self._config.service_dir, XLNetForSequenceClassification
            )
        )
        decoder: ORKGNLPBaseDecoder = TdmExtractorDecoder(labels)
        self._register_pipeline("main", encoder, runner, decoder)

    def __call__(self, text: str, top_n: int = 5) -> Any:
        """
        Extracts Task-Dataset-Metric (TDM) entities from a given
        DocTAET (Title, Abstract, ExperimentalSetup, TableInfo) ``text``

        :param text: `DocTAET <https://doi.org/10.1007/978-3-030-91669-5_35>`_ represented text.
        :param top_n: Top n results to be extracted. Defaults to 5.
        :return: A list of TDMs.
        """
        return self._run(
            raw_input=text,
            top_n=top_n,
            multiple_batches=True,
        )


orkgnlp_context.get("SERVICE_MAP")[TdmExtractor.SERVICE_NAME] = TdmExtractor
